from abc import ABC
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import field
from functools import partial
import logging
from operator import itemgetter

from dime_sampler import DIMEMove
import emcee
import mpmath as mp
import numpy as np
import pandas as pd
from scipy.stats import describe
from scipy.stats import gaussian_kde
import skopt
import xarray as xr

from indica.workflows.bda.priors import ln_prior
from indica.workflows.bda.priors import PriorManager
from indica.workflows.bda.priors import sample_best_half
from indica.workflows.bda.priors import sample_from_priors

exp = np.frompyfunc(mp.exp, 1, 1)
to_float = np.frompyfunc(float, 1, 1)


def exp_wrapper(x: np.ndarray):
    x_star = x * mp.mpf(1)
    y_star = exp(x_star)
    # y = to_float(y_star)
    return np.hstack(y_star[:])


def sample_with_moments(
    sampler,
    start_points,
    iterations,
    n_params,
    auto_sample=10,
    stopping_factor=0.01,
    tune=False,
    debug=False,
):
    # TODO: Compare old_chain to new_chain
    #  if moments are different then keep going / convergence diagnostics here

    autocorr = np.ones(shape=(iterations, n_params)) * np.nan
    old_mean = np.inf
    success_flag = False  # requires succeeding check twice in a row
    for sample in sampler.sample(
        start_points,
        iterations=iterations,
        progress=True,
        skip_initial_state_check=False,
        tune=tune,
    ):
        if sampler.iteration % auto_sample:
            continue
        new_tau = sampler.get_autocorr_time(tol=0)
        autocorr[sampler.iteration - 1] = new_tau

        dist_stats = describe(sampler.get_chain(flat=True))

        new_mean = dist_stats.mean

        dmean = np.abs(new_mean - old_mean)
        rel_dmean = dmean / old_mean

        if debug:
            print("")
            print(f"rel_dmean: {rel_dmean.max()}")
        if rel_dmean.max() < stopping_factor:
            if success_flag:
                break
            else:
                success_flag = True
        else:
            success_flag = False
        old_mean = new_mean

    autocorr = autocorr[
        : sampler.iteration,
    ]
    return autocorr


def sample_from_high_density_region(
    param_names: list, priors: dict, optimiser, nwalkers: int, nsamples=100
):
    num_best_points = 2
    index_best_start = []

    while index_best_start.__len__() < num_best_points:
        start_points = sample_from_priors(param_names, priors, size=nsamples)
        ln_prob, _ = optimiser.compute_log_prob(start_points)
        good_indices = np.argsort(ln_prob)[-num_best_points:]
        index_best_start.extend(good_indices)

    index_best_start[:num_best_points] = index_best_start[:num_best_points]
    best_start_points = start_points[index_best_start, :]
    # Passing samples through ln_prior and redrawing if they fail
    samples = np.empty((param_names.__len__(), 0))
    while samples.size < param_names.__len__() * nwalkers:
        sample = np.random.uniform(
            np.min(best_start_points, axis=0),
            np.max(best_start_points, axis=0),
            size=(nwalkers * 2, len(param_names)),
        )
        start = {name: sample[:, idx] for idx, name in enumerate(param_names)}
        _ln_prior = ln_prior(
            priors,
            start,
        )
        # Convert from dictionary of arrays -> array,
        # then filtering out where ln_prior is -infinity
        accepted_samples = np.array(list(start.values()))[:, _ln_prior != -np.inf]
        samples = np.append(samples, accepted_samples, axis=1)
    start_points = samples[:, 0:nwalkers].transpose()
    return start_points


def bo_wrapper(params: list, dims: list = None, blackbox: callable = None, **kwargs):
    if dims is None:
        raise ValueError("BO wrapper is missing dims")
    if blackbox is None:
        raise ValueError("BO wrapper is missing blackbox")
    params = {dim: params[idx] for idx, dim in enumerate(dims)}
    ln_post, blobs = blackbox({**params}, **kwargs)
    return -ln_post  # probability -> cost function


class OptimiserContext(ABC):
    def __init__(
        self,
    ):
        self.start_points = None
        self.optimiser = None

    @abstractmethod
    def init_optimiser(self, *args, **kwargs):
        self.optimiser = None

    @abstractmethod
    def sample_start_points(self, *args, **kwargs):
        self.start_points = None

    @abstractmethod
    def run(self):
        results = None
        return results

    @abstractmethod
    def post_process_results(self):
        results = {}
        return results

    @abstractmethod
    def reset_optimiser(self):
        self.optimiser = None
        self.start_points = None


@dataclass
class EmceeSettings:
    param_names: list
    iterations: int = 1000
    nwalkers: int = 50
    burn_frac: float = 0.20
    sample_method: str = "random"
    starting_samples: int = 100
    stopping_criteria: str = "mode"
    stopping_criteria_factor: float = 0.01
    stopping_criteria_sample: int = 10
    stopping_criteria_debug: bool = False
    move: list = field(default_factory=lambda: [(DIMEMove(aimh_prob=0.2), 1.0)])


class EmceeOptimiser(OptimiserContext):
    def __init__(
        self,
        optimiser_settings: EmceeSettings,
        prior_manager: PriorManager,
        model_kwargs=None,
    ):
        super().__init__()
        self.optimiser_settings = optimiser_settings
        self.autocorr = None
        self.prior_manager = prior_manager
        self.model_kwargs = model_kwargs
        self.ndim = len(self.optimiser_settings.param_names)

    def init_optimiser(
        self,
        blackbox_func: Callable,
    ):  # type: ignore
        self.optimiser = emcee.EnsembleSampler(
            self.optimiser_settings.nwalkers,
            self.ndim,
            log_prob_fn=blackbox_func,
            kwargs=self.model_kwargs,
            parameter_names=self.optimiser_settings.param_names,
            moves=self.optimiser_settings.move,
        )

    def sample_start_points(
        self,
    ):
        if self.optimiser_settings.sample_method == "high_density":
            self.start_points = sample_from_high_density_region(
                param_names=self.optimiser_settings.param_names,
                priors=self.prior_manager.priors,
                optimiser=self.optimiser,
                nwalkers=self.optimiser_settings.nwalkers,
                nsamples=self.optimiser_settings.starting_samples,
            )
        elif self.optimiser_settings.sample_method == "random":
            self.start_points = sample_from_priors(
                param_names=self.optimiser_settings.param_names,
                priors=self.prior_manager.priors,
                size=self.optimiser_settings.nwalkers,
            )
        else:
            raise ValueError(
                f"Sample method: {self.optimiser_settings.sample_method} "
                f"not recognised, Defaulting to random sampling"
            )

    def run(
        self,
    ):
        if self.optimiser_settings.stopping_criteria == "mode":
            self.autocorr = sample_with_moments(
                self.optimiser,
                self.start_points,
                self.optimiser_settings.iterations,
                self.optimiser_settings.param_names.__len__(),
                auto_sample=self.optimiser_settings.stopping_criteria_sample,
                stopping_factor=self.optimiser_settings.stopping_criteria_factor,
                debug=self.optimiser_settings.stopping_criteria_debug,
            )
        else:
            raise ValueError(
                f"Stopping criteria: "
                f"{self.optimiser_settings.stopping_criteria} invalid"
            )
        return

    def post_process_results(self):
        results = {"convergence": {}}
        _blobs = self.optimiser.get_blobs(
            discard=int(self.optimiser.iteration * self.optimiser_settings.burn_frac),
            flat=True,
        )
        blobs = [blob for blob in _blobs if blob]  # remove empty blobs

        blob_names = blobs[0].keys()
        samples = np.arange(0, blobs.__len__())

        results["blobs"] = {
            blob_name: xr.concat(
                [data[blob_name] for data in blobs],
                dim=pd.Index(samples, name="sample_idx"),
            )
            for blob_name in blob_names
        }
        results["convergence"]["accept_frac"] = self.optimiser.acceptance_fraction.sum()
        results["prior_sample"] = sample_from_priors(
            self.optimiser_settings.param_names,
            self.prior_manager.priors,
            size=int(1e3),
        )

        post_sample = self.optimiser.get_chain(
            discard=int(self.optimiser.iteration * self.optimiser_settings.burn_frac),
            flat=True,
        )
        # pad index dim with maximum number of iterations
        max_iter = self.optimiser_settings.iterations * self.optimiser_settings.nwalkers
        npad = (
            (
                0,
                int(
                    max_iter * (1 - self.optimiser_settings.burn_frac)
                    - post_sample.shape[0]
                ),
            ),
            (0, 0),
        )
        results["post_sample"] = np.pad(post_sample, npad, constant_values=np.nan)
        return results

    def reset_optimiser(self):
        self.optimiser.reset()


@dataclass
class BOSettings:
    param_names: list
    n_calls: int = 30
    acq_func: str = "EI"
    xi: float = 0.01
    kappa: float = 1.96
    n_initial_points: int = 10
    noise: float = 1e-10
    initial_point_generator: str = "lhs"
    use_previous_best: bool = True
    boundary_samples: int = int(1e3)
    model_samples: int = 50
    posterior_samples: int = int(1e5)


class BOOptimiser(OptimiserContext):
    def __init__(
        self,
        optimiser_settings: BOSettings,
        prior_manager: PriorManager,
        model_kwargs: dict = None,
    ):
        super().__init__()
        self.optimiser_settings = optimiser_settings
        if model_kwargs is None:
            model_kwargs = {}
        self.model_kwargs = model_kwargs
        self.wrapped_blackbox_func = None
        self.blackbox_func = None
        self.result = None
        self.prior_manager = prior_manager
        self.ndim = len(self.optimiser_settings.param_names)

        param_samples = sample_from_priors(
            self.optimiser_settings.param_names,
            self.prior_manager.priors,
            size=self.optimiser_settings.boundary_samples,
        )
        self.bounds = [
            (param_samples[:, idx].min(), param_samples[:, idx].max())
            for idx, _ in enumerate(self.optimiser_settings.param_names)
        ]

    def init_optimiser(
        self,
        blackbox_func: Callable,
    ):  # type: ignore
        self.optimiser = skopt.gp_minimize
        self.blackbox_func = partial(blackbox_func, **self.model_kwargs)
        self.wrapped_blackbox_func = partial(
            bo_wrapper,
            dims=self.optimiser_settings.param_names,
            blackbox=blackbox_func,
            **self.model_kwargs,
        )

    def reset_optimiser(self):
        self.optimiser = skopt.gp_minimize
        self.start_points = None

    def sample_start_points(self, *args, **kwargs):
        if self.optimiser_settings.use_previous_best and self.result is not None:
            best_indices = np.argsort(self.result.func_vals)[
                : self.optimiser_settings.n_initial_points
            ]
            self.start_points = list(
                itemgetter(*best_indices)(
                    self.result.x_iters,
                )
            )
        else:
            start_points = sample_best_half(
                self.optimiser_settings.param_names,
                self.prior_manager.priors,
                self.wrapped_blackbox_func,
                self.optimiser_settings.n_initial_points,
            )
            self.start_points = start_points.tolist()

    def run(
        self,
    ):
        self.result = self.optimiser(
            self.wrapped_blackbox_func,
            self.bounds,
            acq_func=self.optimiser_settings.acq_func,
            x0=self.start_points,
            xi=self.optimiser_settings.xi,
            kappa=self.optimiser_settings.kappa,
            n_calls=self.optimiser_settings.n_calls,
            n_initial_points=0,
            noise=self.optimiser_settings.noise,
            acq_optimizer="lbfgs",
            initial_point_generator=self.optimiser_settings.initial_point_generator,
        )

    def post_process_results(
        self,
    ):
        results = {
            "convergence": {"gp_regression": self.result},
        }
        model = self.result.models[-1]

        log = logging.getLogger()
        log.info("kde posterior estimating")

        real_samples = np.array(
            self.result.space.rvs(n_samples=self.optimiser_settings.posterior_samples)
        )
        normed_samples = self.result.space.transform(real_samples)

        obj_func = model.predict(normed_samples)
        posterior = exp_wrapper(
            -obj_func
        )  # log of objective function -> posterior probability
        posterior /= posterior.max()
        posterior = to_float(posterior)
        posterior[posterior == 0] = 1e-100

        posterior_fit = gaussian_kde(real_samples.T, weights=posterior)

        params = posterior_fit.resample(size=self.optimiser_settings.model_samples)

        blobs = []
        for model_sample_idx in range(self.optimiser_settings.model_samples):
            _params = {
                param_name: params[name_idx, model_sample_idx]
                for name_idx, param_name in enumerate(
                    self.optimiser_settings.param_names
                )
            }

            post, _blobs = self.blackbox_func(_params)
            if _blobs:
                blobs.append(_blobs)

        blob_names = blobs[0].keys()
        samples = np.arange(0, blobs.__len__())

        results["blobs"] = {
            blob_name: xr.concat(
                [data[blob_name] for data in blobs],
                dim=pd.Index(samples, name="sample_idx"),
            )
            for blob_name in blob_names
        }
        results["prior_sample"] = sample_from_priors(
            self.optimiser_settings.param_names,
            self.prior_manager.priors,
            size=int(1e3),
        )

        results["post_sample"] = posterior_fit.resample(size=int(2e3)).T
        return results



from typing import Dict

from xarray import DataArray

from indica import Equilibrium
from indica import Plasma
from indica.models.abstract_diagnostic import AbstractDiagnostic


class ModelReader:
    """Reads output of diagnostic forward models"""

    def __init__(
        self,
        models: Dict[str, AbstractDiagnostic],
        model_kwargs: dict = {},
    ):
        """Reader for synthetic diagnostic measurements making use of:

        Parameters
        ----------
        models

        model_kwargs


        """
        self.models = models
        self.model_kwargs = model_kwargs

        self.transforms: dict = {}
        self.plasma = None

        for model_name, model in models.items():
            self.models[model_name] = model(
                name=model_name, **self.model_kwargs.get(model_name, {})
            )

    def set_geometry_transforms(self, transforms: dict, equilibrium: Equilibrium):
        """
        Set instrument geometry and equilibrium
        """

        for instr in self.models.keys():
            if instr not in transforms.keys():
                raise ValueError(f"{instr} not available in given transforms")

            self.transforms[instr] = transforms[instr]
            self.transforms[instr].set_equilibrium(equilibrium, force=True)
            self.models[instr].set_transform(transforms[instr])

    def set_plasma(self, plasma: Plasma):
        """
        Set Plasma class to all models and transforms
        """
        for model_name, model in self.models.items():
            model.set_plasma(plasma)
        self.plasma = plasma

    def update_model_settings(self, **update_kwargs):
        """
        Reinitialise models with model kwargs
        """
        for instrument in self.models:
            self.model_kwargs[instrument] = self.model_kwargs[instrument].update(
                update_kwargs[instrument]
            )
            self.models = self.models[instrument](**self.model_kwargs[instrument])

    def get(
        self,
        instrument: str,
        **kwargs,
    ) -> dict[str, DataArray]:
        """
        Method set to replicate the get() method of the readers
        """
        if instrument in self.models.keys():
            return self.models[instrument](**kwargs)
        else:
            return {}

    def __call__(
        self,
        instruments: list = None,
        **call_kwargs,
    ):
        if instruments is None:
            instruments = self.models.keys()

        bckc: dict = {}
        for instrument in instruments:
            bckc[instrument] = self.get(instrument, **call_kwargs.get(instrument, {}))

        return bckc



"""Coordinate system for data collected on a 1-D along through the Tokamak"""
from typing import Any
from typing import Tuple

import numpy as np
import xarray as xr
from xarray import DataArray

from indica.utilities import format_coord
from .abstractconverter import CoordinateTransform
from ..numpy_typing import Coordinates
from ..numpy_typing import LabeledArray
from ..numpy_typing import OnlyArray


class TransectCoordinates(CoordinateTransform):
    """Coordinate system for measurements taken along a 1-D line through
    the Tokamak.

    The first spatial position corresponds to an arbitrary index along
    the length of the transect. The second is the vertical offset from
    the transect. Typically, all data will fall directly on the
    transect, but it is useful to have the second coordinate so that
    the system is general and to allow conversions to other systems.

    The coordinates along the transect are the indices of the
    positions provided when intialisign the object.

    The default grid for coordinate transforms in this system is a 1-D
    array corresponding to the positions along the axis specified in
    the parmeters used to intialise the object.

    Parameters
    ----------
    x_positions
        1-D array of x-locations along the transect.
    y_positions
        1-D array of y-locations along the transect.
    z_positions
        1-D array of z-locations along the transect.

    """

    def __init__(
        self,
        x_positions: OnlyArray,
        y_positions: OnlyArray,
        z_positions: OnlyArray,
        name: str = "",
        machine_dimensions: Tuple[Tuple[float, float], Tuple[float, float]] = (
            (1.83, 3.9),
            (-1.75, 2.0),
        ),
        **kwargs: Any,
    ):
        if np.shape(x_positions) != np.shape(z_positions) or np.shape(
            y_positions
        ) != np.shape(z_positions):
            raise ValueError("x_, y_ and z_positions must have the same dimensions.")

        self.x1_name: str = "channel"
        self.x2_name: str = ""
        self.instrument_name = name
        self.name: str = f"{name}_transect_transform"

        x1 = np.arange(len(x_positions))
        self.x1: DataArray = format_coord(x1, self.x1_name)
        self.x2: DataArray = DataArray(None)

        # TODO: add intersection with first walls to restrict possible coordinates
        self._machine_dims = machine_dimensions
        self.x = xr.DataArray(
            x_positions,
            coords={self.x1_name: self.x1.data},
        )
        self.y = xr.DataArray(
            y_positions,
            coords={self.x1_name: self.x1.data},
        )
        self.z = xr.DataArray(
            z_positions,
            coords={self.x1_name: self.x1.data},
        )

        self.R: DataArray = np.sqrt(self.x**2 + self.y**2)
        self.rhop: DataArray
        self.theta: DataArray

    def convert_to_Rz(
        self, x1: LabeledArray, x2: LabeledArray, t: LabeledArray
    ) -> Coordinates:
        """Convert from this coordinate to the R-z coordinate system.

        Parameters
        ----------
        x1
            The channel
        x2
            Set to None: not needed in this coordinate transform
        t
            The time coordinate (if there is one, otherwise ``None``)

        Returns
        -------
        R
            Major radius coordinate
        z
            Height coordinate

        """
        R = self.R.sel(channel=x1)
        z = self.z.sel(channel=x1)
        return R, z

    def convert_to_xy(
        self, x1: LabeledArray, x2: LabeledArray, t: LabeledArray
    ) -> Coordinates:
        """Convert from this coordinate to the R-z coordinate system.

        Parameters
        ----------
        x1
            The channel
        x2
            Set to None: not needed in this coordinate transform
        t
            The time coordinate (if there is one, otherwise ``None``)

        Returns
        -------
        x
            x Cartesian coordinate
        y
            y Cartesian coordinate

        """
        x = self.x.sel(channel=x1)
        y = self.y.sel(channel=x1)
        return x, y

    def convert_from_Rz(
        self, R: LabeledArray, z: LabeledArray, t: LabeledArray
    ) -> Coordinates:
        """Convert from the master coordinate system to this coordinate.

        Parameters
        ----------
        R
            Major radius coordinate
        z
            z coordinate
        t
            Time coordinate)

        Returns
        -------
        x1
            The channel
        x2
            Not applicable to this transform: None

        """
        indices = np.where(np.isin(self.R, R) * np.isin(self.z, z))[0]
        x1 = self.x1.isel(channel=indices)
        x2 = DataArray(None)

        return x1, x2

    def map_profile_to_rho(
        self,
        profile_to_map: DataArray,
        t: LabeledArray = None,
        limit_to_sep=True,
        calc_rho=False,
    ) -> DataArray:
        """
        Map profile to measurement coordinates

        Parameters
        ----------
        profile_to_map
            DataArray of the profile to map
        t
            Time for interpolation
        limit_to_sep
            Set to True if values outside of separatrix are to be set to 0
        calc_rho
            Calculate rho for specified time-point

        Returns
        -------
            Interpolation of the input profile on the diagnostic channels
        """
        self.check_equilibrium()
        profile = self.check_rho_and_profile(profile_to_map, t, calc_rho)

        coords = profile.coords
        along_los: DataArray
        if "R" in coords and "z" in coords:
            R_ = self.R
            z_ = self.z

            value_at_channels = profile.interp(R=R_, z=z_).T
        elif "rhop" in coords:
            _rhop = self.rhop
            if "theta" in coords:
                theta_ = self.theta
                value_at_channels = profile.interp(rhop=_rhop, theta=theta_)
            else:
                value_at_channels = profile.interp(rhop=_rhop)

            if limit_to_sep:
                value_at_channels = xr.where(
                    _rhop <= 1,
                    value_at_channels,
                    np.nan,
                )
        else:
            raise NotImplementedError("Coordinates not recognized...")

        drop_coords = [coord for coord in coords if coord != "t"]
        value_at_channels = value_at_channels.drop_vars(drop_coords)
        self.value_at_channels = value_at_channels
        self.profile = profile

        return value_at_channels

    def check_rho_and_profile(
        self, profile_to_map: DataArray, t: LabeledArray = None, calc_rho: bool = False
    ) -> DataArray:
        """
        Check requested time
        """

        time = np.array(t)
        if time.size == 1:
            time = float(time)

        equil_t = self.equilibrium.rhop.t
        equil_ok = (np.min(time) >= np.min(equil_t)) * (np.max(time) <= np.max(equil_t))
        if not equil_ok:
            print(f"Available equilibrium time {np.array(equil_t)}")
            raise ValueError(
                f"Inserted time {time} is not available in Equilibrium object"
            )

        # Make sure rhop.t == requested time
        if not hasattr(self, "rhop") or calc_rho:
            self.convert_to_rho_theta(t=time)
        else:
            if not np.array_equal(self.rhop.t, time):
                self.convert_to_rho_theta(t=time)

        # Check profile
        if not hasattr(profile_to_map, "t"):
            profile = profile_to_map.expand_dims({"t": time})  # type: ignore
        else:
            profile = profile_to_map

        if np.size(time) == 1:
            if np.isclose(profile.t, time, rtol=1.0e-4):
                if "t" in profile_to_map.dims:
                    profile = profile.sel(t=time, method="nearest")
            else:
                raise ValueError("Profile does not include requested time")
        else:
            prof_t = profile.t
            range_ok = (np.min(time) >= np.min(prof_t)) * (
                np.max(time) <= np.max(prof_t)
            )
            if range_ok:
                profile = profile.interp(t=time)
            else:
                raise ValueError("Profile does not include requested time")

        return profile

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._abstract_equals(other)



"""
General equations and quantities for plasma physics
"""

from copy import deepcopy
import math

import numpy as np
from numpy.typing import ArrayLike
import scipy.constants as constants
from scipy.interpolate import CubicSpline
from scipy.interpolate import interp1d


def conductivity_spitzer(el_dens, el_temp, zeff, approx="sauter"):
    """
    Calculate Spitzer conductivity

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Electron temperature (eV)
    zeff
        Effective charge
    approx
        Approximation chosen:
        "sauter" for:
            Sauter et al. Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240
        "todd" for:
            T. Todd et al.
            http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
            doc/TNTodd_Report_NeoclassicalResistivity2003.pdf


    Returns
    -------
    Spitzer conductivity (m**-1 ohm**-1)
    """

    if approx == "sauter":
        return conductivity_spitzer_sauter(el_dens, el_temp, zeff)
    elif approx == "todd":
        return conductivity_spitzer_todd(el_dens, el_temp, zeff)
    else:
        print(f"\n Approximation {approx} not implemented \n")


def conductivity_spitzer_sauter(el_dens, el_temp, zeff):
    """
    Calculate Spitzer conductivity following Sauter et al.
    Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Electron temperature (eV)
    zeff
        Effective charge

    Returns
    -------
    Spitzer conductivity (m**-1 ohm**-1)

    """

    def n_zeff(zeff):
        return 0.58 + 0.74 / (0.76 + zeff)

    coul_log = coul_log_e(el_temp, el_dens)

    conductivity = 1.9012e4 * el_temp ** (3 / 2) / (zeff * n_zeff(zeff) * coul_log)

    return conductivity


def conductivity_spitzer_todd(el_dens, el_temp, zeff):
    """
    Calculate Spitzer conductivity following T. Todd et al.
    http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
    doc/TNTodd_Report_NeoclassicalResistivity2003.pdf

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Electron temperature (eV)
    zeff
        Effective charge

    Returns
    -------
    Spitzer conductivity (m**-1 ohm**-1)

    """

    def lambda_e_zeff(zeff):
        return 3.4 / zeff * (1.13 + zeff) / (2.67 + zeff)

    coul_log = coul_log_e(el_temp, el_dens)
    const = (
        3
        * (2 * math.pi) ** 1.5
        * constants.epsilon_0**2
        / np.sqrt(constants.m_e * constants.e)
    )
    conductivity = const * el_temp ** (3 / 2) * lambda_e_zeff(zeff) / coul_log

    return conductivity


def conductivity_neo(
    el_dens, el_temp, zeff, min_r, a, R_mag, q, vloop=None, approx="sauter"
):
    """
    Calculate Spitzer conductivity

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Electron temperature (eV)
    zeff
        Effective charge
    min_r
        minor radius (m)
    R_mag
        major radius (m)
    q
        safety factor
    vloop
        loop voltage (V)
    approx
        Approximation chosen:
        "sauter" for:
            Sauter et al. Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240
        "todd" for:
            T. Todd et al.
            http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
            doc/TNTodd_Report_NeoclassicalResistivity2003.pdf


    Returns
    -------
    Spitzer conductivity (m**-1 ohm**-1)
    """

    if approx == "sauter":
        return conductivity_neo_sauter(el_dens, el_temp, zeff, min_r, a, R_mag, q)
    elif approx == "todd" and vloop:
        return conductivity_neo_todd(el_dens, el_temp, zeff, min_r, a, R_mag, q, vloop)
    else:
        print(f"\n Approximation {approx} not implemented \n")


def conductivity_neo_sauter(el_dens, el_temp, zeff, min_r, a, R_mag, q):
    """
    Calculate neoclassical conductivity following Sauter et al.
    Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Electron temperature (eV)
    zeff
        Effective charge
    a
        minor radius (m)
    R_mag
        major radius (m)
    q
        safety factor

    Returns
    -------
    Neoclassical conductivity (m**-1 ohm**-1)

    """

    def f33_teff(el_dens, el_temp, zeff, min_r, a, R_mag, q):
        ft = trapped_fraction(min_r, R_mag)
        nu_e = collisionality_electrons_sauter(el_dens, el_temp, zeff, a, R_mag, q)

        return ft / (
            1
            + (0.55 - 0.1 * ft) * np.sqrt(nu_e)
            + 0.45 * (1 - ft) * nu_e / zeff ** (3 / 2)
        )

    def F33(x):
        return (
            1 - (1 + 0.36 / zeff) * x + (0.59 / zeff) * x**2 - (0.23 / zeff) * x**3
        )

    spitzer_cond = conductivity_spitzer_sauter(el_dens, el_temp, zeff)

    x = f33_teff(el_dens, el_temp, zeff, min_r, a, R_mag, q)
    conductivity = spitzer_cond * F33(x)

    return conductivity


def conductivity_neo_todd(el_dens, el_temp, zeff, min_r, a, R_mag, q, vloop):
    """
    Calculate neoclassical conductivity following T. Todd et al.
    http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
    doc/TNTodd_Report_NeoclassicalResistivity2003.pdf

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Electron temperature (eV)
    zeff
        Effective charge
    a
        minor radius (m)
    R_mag
        major radius (m)
    q
        safety factor

    Returns
    -------
    Neoclassical conductivity (m**-1 ohm**-1)

    """

    spitzer_cond = conductivity_spitzer_todd(el_dens, el_temp, zeff)

    ft = trapped_fraction(min_r, R_mag)
    nu_e = collisionality_electrons_todd(el_dens, el_temp, a, R_mag, q, vloop)
    cr = 0.56 / zeff * (3.0 - zeff) / (3.0 - zeff)
    xi = 0.58 + 0.2 * zeff

    conductivity = (
        spitzer_cond * (1 - ft / (1 + xi * nu_e)) * (1 - cr * ft / (1 + xi * nu_e))
    )

    return conductivity


def trapped_fraction(min_r, R_mag):
    """
    Calculate fraction of trapped particles

    To be implemented following T. Todd
    http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
    doc/TNTodd_Report_NeoclassicalResistivity2003.pdf

    or Sauter et al.
    Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240

    Parameters
    ----------
    min_r
        minor radius (m)
    R_mag
        major radius (m)

    Returns
    -------
    Trapped particle fraction

    """
    eps = min_r / R_mag
    ft = np.sqrt(2 * eps / (1 + 2 * eps))

    return ft


def collisionality_electrons_sauter(el_dens, el_temp, zeff, a, R_mag, q):
    """
    Calculate collisionality parameter for electrons following Sauter et al.
    Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240)

    Parameters
    ----------
    el_dens
        Electron density: local (m**-3)
    el_temp
        Electron temperature (eV)
    zeff
        Effective charge
    a
        minor radius
    R_mag
        major radius
    q
        safety factor profile

    Returns
    -------
    Electron collisionality

    """
    coul_log = coul_log_e(el_dens, el_temp)
    epsilon = a / R_mag
    nu = (
        6.921e-18
        * (q * R_mag * el_dens * zeff * coul_log)
        / (el_temp**2 * epsilon ** (3 / 2))
    )

    return nu


def collisionality_electrons_todd(el_dens, el_temp, a, R_mag, q, vloop):
    """
    Calculate collisionality parameter for electrons
    (see T. Todd
    http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
    doc/TNTodd_Report_NeoclassicalResistivity2003.pdf)

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Species temperature (eV)
    R_mag
        major radius (m)
    q
        safety factor profiles
    vloop
        loop voltage (V)

    Returns
    -------
    Electron collisionality

    """

    vthe = thermal_velocity(el_temp, ion=False)
    eps = a / R_mag
    E_perp = constants.m_e * vthe**2 / constants.e
    E_phi = 2 * math.pi * R_mag * vloop
    x = q * R_mag * E_phi / (eps * E_perp)
    tau_ee = collision_time_el_el(el_dens, el_temp)

    nu = (
        1.2
        * np.sqrt(2)
        * R_mag
        * q
        / vthe
        * eps ** (-3 / 2)
        * (1 - x**2) ** (-1 / 4)
        / tau_ee
    )
    return nu


def collision_time_el_el(el_dens, el_temp):
    """
    Calculate electron-electron collision time
    (see T. Todd
    http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
    doc/TNTodd_Report_NeoclassicalResistivity2003.pdf)

    Parameters
    ----------
    el_dens
        Electron density (m**-3)
    el_temp
        Species temperature (eV)

    Returns
    -------
    Electron-electron collision time (s)

    """

    vthe = thermal_velocity(el_temp, ion=False, ev=True)
    coul_log = coul_log_e(el_temp, el_dens)
    tau_ee = (
        3
        / (16 * np.sqrt(math.pi))
        * (4 * math.pi * constants.epsilon_0) ** 2
        * constants.m_e**2
        * vthe**3
        / (el_dens * constants.e**4 * coul_log)
    )

    return tau_ee


def collisionality_ions_sauter(ion_dens, ion_temp, charge, a, R_mag, q):
    """
    Calculate collisionality parameter for ions
    (see Sauter et al. Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240)

    Parameters
    ----------
    ion_dens
        Ion density (m**-3)
    ion_temp
        Ion temperature (eV)
    zeff
        Effective charge
    a
        minor radius
    R_mag
        major radius
    q
        safety factor profiles

    Returns
    -------
    Ion-ion collisionality

    """
    coul_log = coul_log_ii(ion_dens, ion_temp, charge)
    epsilon = a / R_mag
    nu = (
        4.9e-18
        * (q * R_mag * ion_dens * charge**4 * coul_log)
        / (ion_temp**2 * epsilon ** (3 / 2))
    )

    return nu


def coul_log_e(el_temp, el_dens):
    """
    Calculate electron Coulomb logarithm given temperature and density
    (see Sauter et al. Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240)

    Parameters
    ----------
    el_temp
        Electron temperature (eV)
    el_dens
        Electron density (m**-3)

    Returns
    -------
    Coulomb logarithm

    """
    return 31.3 - np.log(np.sqrt(el_dens) / el_temp)


def coul_log_ii(ion_temp, ion_dens, charge):
    """
    Calculate ion-ion Coulomb logarithm given temperature, density and charge
    (see Sauter et al. Physics of Plasmas 6, 2834 (1999); doi: 10.1063/1.873240)

    Parameters
    ----------
    ion_temp
        Ion temperature (eV)
    ion_dens
        Ion density (m**-3)
    charge
        Ion charge (units of electron charge)

    Returns
    -------
    Coulomb logarithm

    """
    return 30 - np.log(charge**3 * np.sqrt(ion_dens) / ion_temp ** (3 / 2))


def current_density(ipla, rho, a, area, prof_shape):
    """
    Build synthetic current density profile (A/m**2) given the
    total plasma current, plasma geometry and a shape parameter

    Parameters
    ----------
    ipla
        total plasma current (A)
    min_r
        minor radius (m)
    a
        separatrix minor radius (m)
    area
        poloidal cross-sectional aread
    nu
        shape parameter

    Returns
    -------

    """
    ir = np.where(rho >= 1)[0]

    j_phi = deepcopy(prof_shape)
    j_phi[ir] = 0.0

    j_phi = ipla * j_phi / np.trapz(j_phi, area)

    return j_phi


def resistance(resistivity, j_phi, area, ipla):
    """
    Calculate the plasma resistance given a current density and resistivity profile

    Parameters
    ----------
    resistivity
        Flux surface averaged resistivity profile (ohm)
    j_phi
        Flux surface averaged current density profile (A/m**2)
    area
        Poloidal cross-section area (m**2)
    ipla
        Current (A)

    Returns
    -------
        Plasma resistance

    """

    res = np.trapz(resistivity * j_phi**2, area) / ipla**2

    return res


def vloop(resistivity, j_phi, area):
    """
    Calculate the loop voltage given the current density, the resistivity
    and geometrical quantitie

    Parameters
    ----------
    resistivity
        Flux surface averaged resistivity profile (ohm)
    j_phi
        Flux surface averaged current density profile (A/m**2)
    area
        Poloidal cross-section area (m**2)
    ipla
        Current (A)

    Returns
    -------
        Plasma resistance

    """

    ipla = np.trapz(j_phi, area)
    vloop = ipla * resistance(resistivity, j_phi, area, ipla)

    return vloop


def poloidal_field(j_phi, min_r, area):
    """
    Calculate the poloidal field given current density and geometry

    Parameters
    ----------
    j_phi
        Toroidal current density (A/m**2)
    min_r
        minor radius (m)
    area
        poloidal cross-section area (m**2)

    Returns
    -------
        Poloidal field (T)

    """

    b_pol = np.zeros_like(j_phi)
    for ir in range(len(min_r)):
        if ir > 0:
            b_pol[ir] = (
                constants.mu_0
                * np.trapz(j_phi[:ir], area[:ir])
                / (2 * math.pi * min_r[ir])
            )

    b_pol[1] = b_pol[2] / 2.0

    return b_pol


def toroidal_field(bt_0, R_mag, maj_r):
    """
    Calculate the toroidal field given vacuum field

    Parameters
    ----------
    bt_0
        Toroidal magnetic field at reference radius R_mag (T)
    R_mag
        Reference major radius
    maj_r
        Major radius (m)

    Returns
    -------
        Toroidal field (T)

    """

    return bt_0 * R_mag / maj_r


def safety_factor(b_tor, b_pol, min_r, a, R_mag, monotonic=True):
    """
    TODO: Calculate safety factor profile following correct formula
    (see T. Todd
    http://mtc-m16.sid.inpe.br/col/sid.inpe.br/marciana/2004/01.15.15.07/
    doc/TNTodd_Report_NeoclassicalResistivity2003.pdf)

    !!!! CURRENTLY JUST APPROXIMATION !!!!

    Parameters
    ----------
    b_tor
        toroidal magnetic field (T)
    b_pol
        poloidal magnetic field (T)
    min_r
        minor radius
    R_mag
        major radius

    Returns
    -------
    safety factor profile

    """
    ir = np.where((min_r < a) * (b_pol > 0))[0]
    q_prof = np.full(min_r.size, np.nan)
    q_prof[ir] = (min_r[ir] * b_tor[ir]) / (R_mag * b_pol[ir])

    if monotonic:
        ir_interp = ir[
            np.where((np.gradient(q_prof[ir]) > 0) * (min_r[ir] > min_r[3]))[0]
        ]
        q_prof[0] = q_prof[np.min(ir_interp)] - 0.05
        ir_interp = np.concatenate([np.array([0]), ir_interp])
        finterp = interp1d(min_r[ir_interp], q_prof[ir_interp], kind="quadratic")
        q_prof[ir] = finterp(min_r[ir])

    return q_prof


def internal_inductance(b_pol, ipla, volume, approx=2, **kwargs):
    """
    Calculate the internal inductance given poloidal magnetic field and
    geometrical quantities

    Parameters
    ----------
    b_pol
        Poloidal magnetic field (T)
    ipla
        Total plasma current (A)
    approx
        Selected approximation

    kwargs
        min_r
            minor radius (m)
        a
            separatrix minor radius (m)
        R_geo
            geometric major radius
        R_mag
            magnetic major radius
        ipla
            total plasma current (A)
        volume
            plasma volume (m**3)
        elongation
            plasma elongation

    Returns
    -------
        Internal inductance l_i

    """

    def li1(b_pol, ipla, volume, min_r, a, R_geo, elongation):
        elongation_a = elongation[np.argmin(np.abs(min_r - a))]
        return (
            2
            * np.trapz(b_pol**2, volume)
            / (R_geo * (constants.mu_0 * ipla) ** 2)
            * (1 + elongation**2 / (2 * elongation_a))
        )

    def li2(b_pol, ipla, volume, R_mag):
        return 2 * np.trapz(b_pol**2, volume) / (R_mag * (constants.mu_0 * ipla) ** 2)

    def li3(b_pol, ipla, volume, R_geo):
        return 2 * np.trapz(b_pol**2, volume) / (R_geo * (constants.mu_0 * ipla) ** 2)

    if (
        approx == 1
        and "min_r" in kwargs.keys()
        and "a" in kwargs.keys()
        and "R_geo" in kwargs.keys()
        and "elongation" in kwargs.keys()
    ):
        return li1(
            b_pol,
            ipla,
            volume,
            kwargs["min_r"],
            kwargs["a"],
            kwargs["R_geo"],
            kwargs["elongation"],
        )

    if approx == 2 and "R_mag" in kwargs.keys():
        return li2(
            b_pol,
            ipla,
            volume,
            kwargs["R_mag"],
        )

    if approx == 3 and "R_geo" in kwargs.keys():
        return li3(
            b_pol,
            ipla,
            volume,
            kwargs["R_geo"],
        )

    return None


def beta(b_field, pressure, volume):
    """
    Calculate ratio of magnetic to kinetic pressure given magnetic field,
    pressure and geometrical quantities

    Parameters
    ----------
    b_field
        Magnetic field (T)
    pressure
        Plasma pressure (Pa)
    volume
        plasma volume

    Returns
    -------
        Beta

    """

    beta = (2 * constants.mu_0 * np.trapz(pressure, volume)) / np.trapz(
        b_field**2, volume
    )

    return beta


def beta_N():
    """
    Calculate normalised beta

    Parameters
    ----------

    Returns
    -------
        Normalised Beta

    """

    return None


def beta_poloidal(b_pol, pressure, volume):
    """
    Calculate the internal inductance given poloidal magnetic field
    and geometrical quantities

    Parameters
    ----------
    b_pol
        Poloidal magnetic field (T)
    pressure
        Plasma pressure (Pa)
    volume
        plasma volume

    Returns
    -------
        Poloidal Beta

    """

    beta_pol = (2 * constants.mu_0 * np.trapz(pressure, volume)) / np.trapz(
        b_pol**2, volume
    )

    return beta_pol


def calc_pressure(density, temperature):
    """
    Calculates pressure in Pa given product

    Parameters
    ----------
    density
        density in units of (m**-3)
    temperature
        temperature in units of (eV)

    Returns
    -------
    Pressure in Pascal

    """

    return density * temperature * constants.e


def thermal_velocity(temperature, mass=1, ion=False, ev=False):
    """
    Calculates the thermal velocity given temperature (eV) and mass (AMU)

    Parameters
    ----------
    temperature
        Species temperature in (eV)
    mass
        Species mass (AMU)
    ion
        If False: species = electron;
        if True: species is an ion
    ev
        If True: temperature is kept in eV and not converted to K

    Returns
    -------
    Thermal velocity (m/s)

    """
    ev2k = constants.physical_constants["electron volt-kelvin relationship"][0]
    k_B = constants.k
    if ion is False:
        amu2kg = constants.m_e
    else:
        amu2kg = constants.m_p

    if ev is False:
        vth = np.sqrt(temperature * ev2k * k_B / (mass * amu2kg))
    else:
        vth = np.sqrt(temperature / (mass * amu2kg))

    return vth


def doppler_ev(sigma, centroid, mass: float, sigma_instr=0.0, fwhm=False):
    """
    Convert spectral line width to a temperature (eV)

    Parameters
    ----------
    sigma
        Gaussian sigma
    centroid
        Gaussian centroid
    mass
        Atomic number of the emitting element
    sigma_instr
        Instrument function sigma
    fwhm
        Set to true if input sigmas are given as Full Width at Half Maximum
        with fwhm = sigma * 2 * np.sqrt(2 * np.log(2))

    Returns
    -------
    temperature
        Temperature of the species in eV

    """

    sigma_thermal = np.sqrt(sigma**2 - sigma_instr**2)

    J2eV = constants.physical_constants["joule-electron volt relationship"][0]
    temperature = (
        mass
        * constants.m_p
        * J2eV
        * constants.c**2
        * (sigma_thermal / centroid) ** 2.0
    )

    if fwhm is True:
        temperature /= 8 * np.log(2)

    return temperature


def ev_doppler(temperature, mass: float, fwhm=False):
    """
    Convert spectral line width to a temperature (eV)

    Parameters
    ----------
    temperature
        Temperature of the ion species in eV
    mass
        Atomic number of the emitting element
    fwhm
        Set to true if input sigma is to given as a Full Width at Half Maximum
        with fwhm = sigma * 2 * np.sqrt(2 * np.log(2))

    Returns
    -------
    dl_l
        Relative width of the spectral line expressed as sigma/centroid

    """

    J2eV = constants.physical_constants["joule-electron volt relationship"][0]
    dl_l = np.sqrt(temperature / (mass * constants.m_p * J2eV * constants.c**2))

    if fwhm is True:
        dl_l *= 2 * np.sqrt(2 * np.log(2))

    return dl_l


def centrifugal_asymmetry(
    ion_temperature,
    electron_temperature,
    mass,
    meanz,
    zeff,
    main_ion_mass,
    toroidal_rotation=None,
    asymmetry_parameter=None,
):

    """
    Calculate toroidal rotation or asymmetry parameter for given plasma parameters


    Parameters
    ----------
    ion_temperature
        Ion temperature (eV)
    electron_temperature
        Electron temperature (eV)
    mass
        Atomic mass of ion whose centrifugal asymmetry is to be investigated
    meanz
        Average charge of ion whose centrifugal asymmetry is to be investigated
    zeff
        Plasma effective charge
    toroidal_rotation
        Toroidal rotation (rad/s)
    asymmetry_parameter

    Returns
    -------
    toroidal_rotation or asymmetry_parameter following equations in
        J. A. Wesson 1997 Nucl. Fusion 37 577
    TODO: include simple equations for fast particle drive asymmetry given in
        T. Odstrcil et al 2018 Plasma Phys. Control. Fusion 60 014003

    """

    const = (mass * constants.proton_mass) / (2 * ion_temperature * constants.e)
    const *= 1 - (meanz / mass) * (main_ion_mass * zeff * electron_temperature) / (
        ion_temperature + zeff * electron_temperature
    )

    if toroidal_rotation is not None:
        asymmetry_parameter = const * toroidal_rotation**2
        return asymmetry_parameter
    elif asymmetry_parameter is not None:
        toroidal_rotation = np.sqrt(asymmetry_parameter / const)
        return toroidal_rotation
    else:
        print(
            "\n physics.centrifugal_asymmetry: input either toroidal_rotation "
            "or asymmetry parameter \n"
        )
        raise ValueError


def zeff_bremsstrahlung(
    Te: ArrayLike,
    Ne: ArrayLike,
    wavelength: float,
    zeff: ArrayLike = None,
    bremsstrahlung: ArrayLike = None,
    gaunt_approx="callahan",
):
    """
    Calculate Bremsstrahlung <-> Zeff using the formulas from:
    "Zeff Measurement and Analysis of Reheat-Mode Discharges in W7-X"
    S. Morita , IPP III / 199 , 1994
    https://core.ac.uk/download/pdf/210800245.pdf

    Parameters
    ----------
    Te
        electron temperature (eV) for calculation
    Ne
        electron density in (m^-3)
    wavelength
        filter central wavelength (nm)
    zeff
        effective charge
    bremsstrahlung
        Bremsstrahlung emission in units [W / m^3 / nm]
    gaunt_approx
        approximation for free-free gaunt factors:
            "callahan" see KJ Callahan 2019 JINST 14 C10002
            "carson" see TR Carson 1988 Atron. Atrophys. 1889, 319324

    Returns
    -------
    Zeff
        or
    Bremsstrahlung [W / m^3 / nm]

    N.B. Comparison with experimental Bremsstrahlung spectrally-resolved
    measurement requires LOS-integral & multiplication by the system's
    Etendue / 4 / np.pi. For filtered diode, the result should also be
    multiplied by the transmission curve and integrated over wavelength.
    """

    assert (zeff is not None) == (bremsstrahlung is None)

    wavelength_ang = wavelength * 10  # nm to Angstrom

    gaunt_funct = {
        "callahan": 1.35 * Te**0.15,
        "carson": -4.7499 + 0.5513 * np.log(Te * wavelength_ang),
    }
    gaunt = gaunt_funct[gaunt_approx]

    Ne_cm = Ne * 1e-6  # m^-3 to cm^-3

    reconvert_units = 1e6 * 10  # convert back to (nm) and (m**-3)

    factor = (
        1.89e-28
        * gaunt
        * Ne_cm**2
        / (np.sqrt(Te) * wavelength_ang**2)
        * np.exp(-12400 / (wavelength_ang * Te))
    ) * reconvert_units

    if zeff is None:
        result = bremsstrahlung / factor
    else:
        result = zeff * factor

    return result


def nm_eV_conversion(nm=None, ev=None):
    if nm is None and ev is None:
        return None
    if nm is not None and ev is not None:
        raise Exception("Input either nm or eV, not both")

    nm_to_m = 1.0e-9
    const = constants.h * constants.c / constants.e / nm_to_m

    if ev is None:
        result = const / nm
    else:
        result = const / ev

    return result


def derivative(y, x):
    nlen = len(x)
    der = np.zeros(nlen)
    for i in range(nlen - 1):
        if i > 0:
            der[i] = (y[i + 1] - y[i - 1]) / (x[i + 1] - x[i - 1])

    finterp = interp1d(x[1 : nlen - 1], y[1 : nlen - 1], kind="quadratic")
    der = finterp(x)

    return der


def gaussian(x, integral, center, sigma):
    return (
        integral
        / (sigma * np.sqrt(2 * np.pi))
        * np.exp(-((x - center) ** 2) / (2 * sigma**2))
    )


def doppler_broaden(x, integral, center, ion_mass, ion_temp):
    _mass = ion_mass * constants.proton_mass * constants.c**2
    sigma = np.sqrt(constants.e / _mass * ion_temp) * center
    gaussian_broadened = gaussian(
        x,
        integral,
        center,
        sigma,
    )
    return gaussian_broadened


def make_window(
    x: ArrayLike,
    x_center: float,
    fwhm: float,
    amplitude: float = 1.0,
    background: ArrayLike = 0.0,
    window: str = "gaussian",
):
    """
    Build a window with known parameters

    Parameters
    ----------
    x
        abscissa
    amplitude
        peak from background
    background
        background value
    x_center
        center of peak
    fwhm
        full with at half maximum
    window
        type of window
    """
    sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))
    windows = {
        "gaussian": np.exp(-((x - x_center) ** 2) / (2 * sigma**2)),
        "boxcar": (
            np.heaviside(x - (x_center - fwhm / 2), 1)
            - np.heaviside(x - (x_center + fwhm / 2), 1)
        ),
    }

    return (amplitude - background) * windows[window] + background


def sawtooth_crash(xspl, yspl, volume, x_inv):
    vol_int_pre = np.trapz(yspl, volume)
    inv_ind = np.max(np.where(xspl <= x_inv)[0])
    for xind in np.arange(inv_ind, xspl.size):
        yspl = np.where(xspl <= xspl[xind], yspl[inv_ind], yspl)
        vol_int_post = np.trapz(yspl, volume)
        if vol_int_post >= vol_int_pre:
            break

    yspl[xind : xind + 5] = np.linspace(yspl[xind], yspl[xind + 5], 5)
    x = np.linspace(0, 1, int(np.size(xspl) / 1.2))

    y = np.interp(x, xspl, yspl)
    cubicspline = CubicSpline(
        x,
        y,
        0,
        "clamped",
        False,
    )
    yspl = cubicspline(xspl)
    vol_int_post = np.trapz(yspl, volume)
    print(f"Vol-int: {float(vol_int_pre)}, {float(vol_int_post)}")

    return yspl


def calc_moments(y: np.array, x: np.array, ind_in=None, ind_out=None, simmetry=False):
    x_avrg = np.nansum(y * x) / np.nansum(y)

    if (ind_in is None) and (ind_out is None):
        ind_in = x <= x_avrg
        ind_out = x >= x_avrg
        if simmetry is True:
            ind_in = ind_in + ind_out
            ind_out = ind_in

    x_err_in = np.sqrt(
        np.nansum(y[ind_in] * (x[ind_in] - x_avrg) ** 2) / np.nansum(y[ind_in])
    )

    x_err_out = np.sqrt(
        np.nansum(y[ind_out] * (x[ind_out] - x_avrg) ** 2) / np.nansum(y[ind_out])
    )

    return x_avrg, x_err_in, x_err_out, ind_in, ind_out



import time as tt

import matplotlib.pylab as plt
from matplotlib.widgets import Button
from matplotlib.widgets import Slider
import numpy as np
from scipy.interpolate import interp1d
from scipy.interpolate import RectBivariateSpline
from scipy.linalg import eigh
from scipy.linalg import solve_banded
from scipy.ndimage import convolve
from scipy.signal import argrelextrema

# mpl.rcParams['keymap.back'].remove('left')
# mpl.rcParams['keymap.forward'].remove('right')


def update_fill_between(fill, x, y_low, y_up, min, max):
    (paths,) = fill.get_paths()
    nx = len(x)

    y_low = np.maximum(y_low, min)
    y_low[y_low == max] = min
    y_up = np.minimum(y_up, max)

    vertices = paths.vertices.T
    vertices[:, 1 : nx + 1] = x, y_up
    vertices[:, nx + 1] = x[-1], y_up[-1]
    vertices[:, nx + 2 : -1] = x[::-1], y_low[::-1]
    vertices[:, 0] = x[0], y_up[0]
    vertices[:, -1] = x[0], y_up[0]


def update_errorbar(err_plot, x, y, yerr):

    plotline, caplines, barlinecols = err_plot

    # Replot the data first
    plotline.set_data(x, y)

    # Find the ending points of the errorbars
    error_positions = (x, y - yerr), (x, y + yerr)

    # Update the caplines
    if len(caplines) > 0:
        for j, pos in enumerate(error_positions):
            caplines[j].set_data(pos)

    # Update the error bars
    barlinecols[0].set_segments(
        list(zip(list(zip(x, y - yerr)), list(zip(x, y + yerr))))
    )


def triband_transpose(A):
    # transpose the tridiagonal band matrix A
    AT = np.copy(A)
    AT[0, 1:], AT[2, :-1] = A[2, :-1], A[0, 1:]
    return AT


def triband_diag_multi(A, D):
    # multiply tridiagonal band matrix a diagonal matrix

    AD = np.copy(A)
    AD[0, 1:] *= D[:-1]
    AD[1] *= D
    AD[2, :-1] *= D[1:]
    return AD


def fast_svd(H):
    # fast method to calculate U,S,V = svd(H.T) of thin rectangular matrix
    LL = np.dot(H.T, H)
    S2, U = eigh(LL, overwrite_a=True, check_finite=True, lower=True)
    S = (
        np.maximum(S2, S2[-1] * 1e-20) ** 0.5
    )  # singular values S can be negative due to numerical uncertainty
    return U, S


class SXR_tomography:
    def __init__(
        self,
        input_dict: dict,
        reg_level_guess: float = 0.5,
        reg_level_min: float = 0.4,
        reg_level_max=0.4,
        nfisher=3,
        eps=1e-5,
        optim_regul: bool = False,
    ):
        """
        1D tomography code assuming the emissivity is flux-surface symmetric

        Parameters
        ----------
        input_dict
            input_dict = dict(
                brightness:array(ntime, nchannel) = measured experimental brightness,
                dl:float = radial precision of the lines of sight,
                t:array(ntime) = time array of the experimental data,
                R:array(nchannel, npoints) = R along the LOS (npoints along LOS),
                z:array(nchannel, npoints) = z along the LOS (npoints along LOS),
                rho_equil=dict(
                    R:array(nR_eq) = R grid of the equilibrium,
                    z:array(nz_eq) = z grid of the equilibrium,
                    t:array(nt_eq) = time of the equilibrium,
                    rho:array(nt_eq, nR_eq, nz_eq) = sqrt(normalised pol flux),
                ),
                debug=debug,
                has_data:array(nchannel) = boolean array to discard non-active channels,
            )
        reg_level_guess:float = regularisation (larger value --> stiffer profiles)
        """

        self.reg_level_guess = reg_level_guess
        self.reg_level_min = reg_level_min
        self.reg_level_max = reg_level_max
        self.nfisher = nfisher
        self.eps = eps
        self.optim_regul = optim_regul

        self.eq = input_dict["rho_equil"]

        # Parameters
        los_shape = np.shape(input_dict["R"])
        self.nlos = los_shape[0]
        self.los_len = los_shape[1]

        # number of equalibrium timeslices
        self.nt = self.eq["t"].size

        # number of radial grid binds
        self.nr = 100

        # number of virtual LOS accounting for finite divergence of LOS
        self.nvirt = 11  # odd number

        # radial grid
        self.rho_grid_edges = np.linspace(0, 1, self.nr + 1)
        self.rho_grid_centers = (self.rho_grid_edges[1:] + self.rho_grid_edges[:-1]) / 2

        # step along LOS
        self.dL = input_dict["dl"]

        self.R = input_dict["R"]
        self.z = input_dict["z"]
        # self.rho = input_dict['rho']  # TODO: channel list

        # IMPACT PARAMETERS
        # self.impact_parameters = input_dict["impact_parameters"]

        self.debug = input_dict["debug"]

        # load data
        self.data = input_dict["brightness"]  # W/m^2 ??

        # load error
        default_err = self.data * 0.05 + np.nanmax(self.data) * 0.01  # assume 5% error
        if "brightness_error" in input_dict.keys():
            self.err = input_dict["brightness_error"]  # W/m^2 ??
        else:
            self.err = default_err

        # guess!! of uncertainty
        if np.all(self.err) == 0:
            self.err = default_err

        self.tvec = input_dict["t"]

        self.valid = input_dict["has_data"]

        if "emissivity" in input_dict.keys():
            self.expected_emissivity = input_dict["emissivity"]

    def geom_matrix(self):
        # create geometry matrix witn contribution of all
        # grid intervals to the measured signals

        # create vitual LOS
        self.x = np.arange(self.nlos)
        self.x_inter = (
            self.x[:, None] + np.linspace(-1, 1, self.nvirt) * (1 - 1 / self.nvirt) / 2
        )

        R = interp1d(self.x, self.R.T, fill_value="extrapolate")(self.x_inter)
        z = interp1d(self.x, self.z.T, fill_value="extrapolate")(self.x_inter)

        # locate index x_tg for R_tg
        i_tg = np.argmin(R, axis=0)[None]

        # weighted sum all dL values which fits in between bin edges
        # dLmat2 is calculate by summing all dL contributiions to
        # each grid bin, used for benchmarking
        nt = self.eq["t"].size
        # calculate exactly length of chord in each grid bin
        dLmat = np.zeros((nt, self.nlos, self.nvirt, self.nr))
        rho_tg = np.zeros((nt, self.nlos, self.nvirt))

        #  prepare L matrix
        for it in range(nt):
            LOS_rho = RectBivariateSpline(
                self.eq["R"], self.eq["z"], self.eq["rho"][it].T
            ).ev(R, z)

            rho_tg[it] = np.take_along_axis(LOS_rho, i_tg, axis=0)[0]

            # iterate over chords and calculate contribution to each grid bin
            for ilos in range(self.nlos):
                # weight is given by dL value and is is splitted equally
                # between all nvirt virtual LOSs

                for iv in range(self.nvirt):
                    # get first and last point just outside of lcfs
                    ind = np.where(np.abs(LOS_rho[:, ilos, iv]) < 1)[0]
                    if len(ind) == 0:  # no cross section with grid
                        continue
                    ilcfs_in = ind[0] - 1
                    ilcfs_out = ind[-1] + 2
                    rho_cut = LOS_rho[ilcfs_in:ilcfs_out, ilos, iv]

                    # find local minima and maxima, rho is monotonous in between
                    imin = argrelextrema(rho_cut, np.less_equal)[0]
                    imax = argrelextrema(rho_cut, np.greater_equal)[0]
                    i_extrema = np.unique(np.r_[imin, imax])

                    L = np.arange(ilcfs_out - ilcfs_in) * self.dL
                    L_turn = 0

                    for i in range(len(i_extrema) - 1):
                        # split LOS in regions with monotonously changing
                        # rho to make the inversion
                        monotone_ind = slice(i_extrema[i], i_extrema[i + 1] + 1)

                        if rho_cut[monotone_ind][0] == rho_cut[monotone_ind][-1]:
                            # special case, whole monotone_ind is constant
                            continue

                        # input for interpolation needs to be monotonous
                        Ledge = interp1d(
                            rho_cut[monotone_ind],
                            L[monotone_ind],
                            bounds_error=False,
                            fill_value=0,
                        )(self.rho_grid_edges)

                        if not np.any(Ledge > 0):
                            # special case, whole monotone_ind is
                            # within single grid cell
                            continue

                        # index of turning points
                        i_tg_in, i_tg_out = np.where(Ledge > 0)[0][[0, -1]]

                        dLmat[it, ilos, iv, i_tg_in:i_tg_out] += np.abs(
                            Ledge[i_tg_in:i_tg_out] - Ledge[i_tg_in + 1 : i_tg_out + 1]
                        )

                        # add the turning point singularity
                        if (
                            rho_cut[monotone_ind][-1] - rho_cut[monotone_ind][0] > 0
                        ):  # if L increases with rho
                            if i_tg_in > 0:
                                dLmat[it, ilos, iv, i_tg_in - 1] += (
                                    Ledge[i_tg_in] - L_turn
                                )
                            if (
                                i_tg_out < self.nr
                            ):  # this should happen just in the last step
                                L_turn = Ledge[i_tg_out]
                                assert i < len(i_extrema) - 1
                        else:  # if L decreases with rho
                            if (
                                i_tg_out < self.nr
                            ):  # this should happen just in the first step
                                dLmat[it, ilos, iv, i_tg_out] += (
                                    Ledge[i_tg_out] - L_turn
                                )
                            L_turn = Ledge[i_tg_in]

        # average over all virtual LOSs, assume equal weight
        self.dLmat = dLmat.mean(2)
        self.rho_tg = rho_tg.mean(2)

        dLmat_interleaved = dLmat.reshape(nt, self.nvirt * self.nlos, self.nr)
        self.dLmat_interleaved = convolve(
            dLmat_interleaved, np.ones((1, self.nvirt, 1)) / self.nvirt, mode="nearest"
        )

        rho_tg_interleaved = rho_tg.reshape(nt, -1)
        self.rho_tg_interleaved = convolve(
            rho_tg_interleaved, np.ones((1, self.nvirt)) / self.nvirt, mode="nearest"
        )

    def regul_matrix(self, bias_axis=True, bias_edge=False):
        # regularization band matrix, 2. order derivative,
        # bias left or right side to zero
        bias = 0.1
        D = np.ones((3, self.nr))
        D[1, :] *= -2
        D[0, 1] = 0
        D[2, -2] = 0

        D[1, [0, -1]] = 1e-5  # just to make D invertible

        if bias_axis:
            # bias to zero gradient on axis
            D[1, 0] = 1
            D[0, 1] = -1

        if bias_edge:
            # bias right edge value towards zero
            D[1, -1] = bias

        return D

    def PRESS(self, g, prod, S, U):
        # predictive sum of squares
        w = 1.0 / (1.0 + np.exp(g) / S**2)
        ndets = len(prod)
        return (
            np.sum(
                (np.dot(U, (1 - w) * prod) / np.einsum("ij,ij,j->i", U, U, 1 - w)) ** 2
            )
            / ndets
        )

    def GCV(self, g, prod, S, U):
        # generalized crossvalidation
        w = 1.0 / (1.0 + np.exp(g) / S**2)
        ndets = len(prod)
        return (np.sum((((w - 1) * prod)) ** 2) + 1) / ndets / (1 - np.mean(w)) ** 2

    def FindMin(self, F, x0, dx0, prod, S, U, tol=0.01):
        # stupid but robust minimum searching algorithm.

        fg = F(x0, prod, S, U)
        while abs(dx0) > tol:
            fg2 = F(x0 + dx0, prod, S, U)

            if fg2 < fg:
                fg = fg2
                x0 += dx0
                continue
            else:
                dx0 /= -2.0

        return x0, np.log(fg2)

    def calc_tomo(self, reg_level=0.8, nfisher=3, eps=1e-5, optim_regul=False):
        # calculate tomography using optimised minimum Fisher regularisation
        # Odstrcil, T., et al. "Optimized tomography methods for plasma
        # emissivity reconstruction at the ASDEX  Upgrade tokamak.
        # Review of Scientific Instruments 87.12 (2016): 123505.

        # prepare regularisation operator - contains all prior information
        # biased_edges - assume zero value at the boundaries of the grid.
        D = self.regul_matrix()

        nt = len(self.tvec)

        self.chi2 = np.zeros(nt)
        self.gamma = np.zeros(nt)

        self.backprojection_int = np.zeros((nt, self.nlos * self.nvirt))
        self.backprojection = np.ones((nt, self.nlos)) * np.nan
        self.emiss = np.zeros((nt, self.nr))
        self.emiss_err = np.zeros((nt, self.nr))

        teq_ind = np.int_(
            interp1d(
                self.eq["t"],
                np.arange(len(self.eq["t"])),
                kind="nearest",
                fill_value="extrapolate",
            )(self.tvec)
        )

        # split reconstruction in blocks nearest to equilibrium data
        for it, teq in enumerate(self.eq["t"]):
            t_ind = teq_ind == it
            if not np.any(t_ind):
                continue

            valid = (
                np.all(
                    (np.isfinite(self.data[t_ind, :])) & (self.data[t_ind, :] > 0),
                    axis=0,
                )
                * self.valid
            )

            # weight the contribution matrix and data by the uncertainty
            # print(it,teq,t_ind)#,(self.err[t_ind,:]/self.data[t_ind,:]).shape,valid)
            err = np.atleast_2d(self.err[t_ind, valid]).mean(0)
            T = self.dLmat[it, valid] / err[:, None]
            mean_d = (
                np.atleast_2d(self.data[t_ind, valid]).mean(0) / err
            )  # NOTE take a ratio of means of mean of rations?
            d = np.atleast_2d(self.data[t_ind, valid] / self.err[t_ind, valid])

            # flat initial estimate of the weight matrix W
            W = np.ones(self.nr)

            q = np.linspace(0, 1, len(mean_d))

            # iterative calculation of minimum Fisher regularisation
            for ifisher in range(nfisher):

                # solve Tikhonov regularization (optimised for speed)
                # multiply tridiagonal regularisation operator by a
                # diagonal weight matrix W

                WD = triband_diag_multi(D, W**0.5)

                # transpose the band matrix WD
                DTW = triband_transpose(WD)

                # calculate (D.TW)^-1*T.T
                H = solve_banded(
                    (1, 1), DTW, T.T, overwrite_ab=False, check_finite=False
                )

                invalid = H.sum(0) == 0
                if np.any(invalid):
                    print("Warning - some LOS are not linearly independent")

                # fast method to calculate U,S,V = svd(H.T) of rectangular matrix
                U, S = fast_svd(H)

                # projection of the data on the U base
                mean_p = np.dot(mean_d, U)

                # calculate regularisation parameter
                g0 = np.interp(reg_level, q, 2 * np.log(S))

                # filtering factors attenuating high frequency eigenvectors
                w = 1.0 / (1.0 + np.exp(g0) / S**2)

                # mean solution
                y = np.dot(H, np.dot(U / S**2, w * mean_p))
                # final inversion of  solution, reconstruction
                y = solve_banded(
                    (1, 1),
                    WD,
                    y,
                    overwrite_ab=False,
                    overwrite_b=True,
                    check_finite=False,
                )

                W = 1 / np.maximum(y, eps)

            # evaluate the basis in the reconstruction space
            V = np.dot(H, U / S)
            V = solve_banded(
                (1, 1), WD, V, overwrite_ab=False, overwrite_b=True, check_finite=False
            )

            # estimate optimal regularisation level for initial guess
            if optim_regul:
                g0, log_fg2 = self.FindMin(self.PRESS, g0, 1, mean_p, S, U)
                self.gamma[t_ind] = np.interp(g0, np.log(S) * 2, q)
                self.gamma[t_ind] = max(self.reg_level_min, self.gamma[t_ind]) ** 2
                w = 1.0 / (1.0 + np.exp(g0) / S**2)
            else:
                self.gamma[t_ind] = reg_level

            # find solution for each timeslice
            p = np.dot(d, U)
            y = np.dot((w / S) * p, V.T)
            fit = np.dot(p * w, U.T)

            self.backprojection[t_ind, valid] = fit * err
            # self.backprojection[t_ind,valid] = np.dot(y, self.dLmat[it].T)[:,valid]

            self.backprojection_int[t_ind] = np.dot(y, self.dLmat_interleaved[it].T)

            self.chi2[t_ind] = np.sum((d - fit) ** 2, 1) / np.size(
                fit, 1
            )  # TODO jen ty funkcni ch

            self.emiss[t_ind] = y
            self.emiss_err[t_ind] = np.sqrt(
                np.dot(V**2, (w / S) ** 2) * np.maximum(1, self.chi2[t_ind, None])
            )

    def show_reconstruction(self):
        # reg_level_min = .4
        # reg_level_max = .4
        f, ax = plt.subplots(1, 3, sharex="col", figsize=(12, 5))
        ax_reg = plt.axes([0.2, 0.02, 0.65, 0.03], facecolor="y")
        slide_reg = Slider(
            ax_reg,
            "Regularisation:",
            0,
            1,
            valinit=self.reg_level_guess,
            valstep=0.001,
            valfmt="%1.3f",
        )

        ax_time = plt.axes([0.2, 0.07, 0.65, 0.03], facecolor="y")
        slide_time = Slider(
            ax_time,
            "Time [s]:",
            self.tvec[0],
            self.tvec[-1],
            valinit=self.tvec[0],
            valstep=0.001,
            valfmt="%1.3f",
        )

        r = self.rho_grid_centers
        f.subplots_adjust(bottom=0.2)

        tomo_var = ax[0].fill_between(
            r, r * 0, r * 0, alpha=0.5, facecolor="b", edgecolor="None"
        )
        (expected_emissivity,) = ax[0].plot([], [], color="r", linestyle="dashed", lw=2)
        (tomo_mean,) = ax[0].plot([], [], lw=2)

        errorbar = ax[1].errorbar(
            0, np.nan, 0, capsize=4, c="r", marker="o", ls="none", alpha=0.5
        )
        (retro_inter,) = ax[1].plot([], [], "b-")
        (retro,) = ax[1].plot([], [], "bx")

        ax[0].axhline(0, c="k")
        ax[1].axhline(0, c="k")

        ax[0].set_xlim(0, 1)
        ax[1].set_xlim(self.x_inter.min(), self.x_inter.max())

        ax[1].set_ylim(0, np.nanmax(self.data) * 1.2)
        ymax = np.nanmax(self.emiss)
        if hasattr(self, "expected_emissivity"):
            ymax = np.nanmax([ymax, self.expected_emissivity.max()])
        ax[0].set_ylim(0, ymax * 1.2)

        ax[0].set_xlabel(r"$\rho$")
        ax[1].set_xlabel(r"Channel")
        ax[1].set_ylabel("Brightness [W/m$^2$]")
        ax[0].set_ylabel("Emissivity [W/m$^3$]")
        ax[0].ticklabel_format(style="sci", axis="y", scilimits=(0, 0))
        ax[1].ticklabel_format(style="sci", axis="y", scilimits=(0, 0))

        global cont
        cvals = np.linspace(0, 1, 20)

        cont = ax[2].contour(
            self.eq["R"], self.eq["z"], self.eq["rho"][0], cvals, colors="red"
        )

        ax[2].axis("equal")
        ax[2].plot(self.R.T, self.z.T, "b", zorder=99)
        ax[2].axis(
            [
                0.15,
                0.7,
                -0.4,
                0.4,
            ]
        )
        ax[2].set_ylabel("z [m]")
        ax[2].set_xlabel("R [m]")
        ax[2].set_xlim(self.eq["R"][[0, -1]])
        ax[2].set_ylim(self.eq["z"][[0, -1]])

        f.subplots_adjust(wspace=0.3)

        title = f.suptitle("")

        def update(reg=None, time=None):
            global cont
            if reg is not None:
                self.calc_tomo(reg_level=reg, nfisher=3, eps=1e-5)

            if time is None:
                time = slide_time.val

            it = np.argmin(np.abs(self.tvec - time))

            update_fill_between(
                tomo_var,
                r,
                self.emiss[it] - self.emiss_err[it],
                self.emiss[it] + self.emiss_err[it],
                -np.inf,
                np.inf,
            )
            if hasattr(self, "expected_emissivity"):
                expected_emissivity.set_data(
                    self.expected_emissivity.rhop,
                    self.expected_emissivity.sel(t=time, method="nearest"),
                )
            tomo_mean.set_data(r, self.emiss[it])

            retro_inter.set_data(self.x_inter, self.backprojection_int[it])
            retro.set_data(self.x, self.backprojection[it])

            update_errorbar(errorbar, self.x, self.data[it], self.err[it])

            for c in cont.collections:
                c.remove()  # removes only the contours, leaves the rest intact

            it_eq = np.argmin(np.abs(self.eq["t"] - time))

            cont = ax[2].contour(
                self.eq["R"],
                self.eq["z"],
                self.eq["rho"][it_eq],
                cvals,
                colors=["r"] * len(cvals),
                alpha=0.8,
            )

            title.set_text(
                r"  $\chi^2/nDoF$ = " + f"{self.chi2[it]:.1f}  "
                r" $\gamma$ = " + f"{self.gamma[it]:.2f}"
            )
            f.canvas.draw_idle()

        def update_time(time):
            update(time=time)

        def update_reg(reg):
            update(reg=reg)

        def on_key(event):
            dt = (self.tvec[-1] - self.tvec[0]) / (len(self.tvec) - 1)
            tnew = slide_time.val

            if hasattr(event, "step"):
                # scroll_event
                tnew += event.step * dt

            elif "left" == event.key:
                # key_press_event
                tnew -= dt

            elif "right" == event.key:
                tnew += dt

            tnew = min(max(tnew, self.tvec[0]), self.tvec[-1])
            slide_time.set_val(tnew)
            update(time=tnew)

        slide_reg.on_changed(update_reg)
        slide_time.on_changed(update_time)

        update(slide_reg.valinit, slide_time.valinit)

        self.cid = f.canvas.mpl_connect("key_press_event", on_key)
        self.cid_scroll = f.canvas.mpl_connect("scroll_event", on_key)

        axbutton = plt.axes([0.85, 0.1, 0.1, 0.05])
        self.save_button = Button(axbutton, "Save")
        self.save_button.on_clicked(self.save)
        plt.show()

    def save(self, args):
        np.savez_compressed(
            "Reconstruction",
            backprojection=self.backprojection,
            brightness=self.data,
            brightness_err=self.err,
            time=self.tvec,
            grid=self.rho_grid_centers,
            gamma=self.gamma,
            chi2=self.chi2,
            emiss=self.emiss.astype("single"),
            emiss_err=self.emiss_err.astype("single"),
        )

        print("Saved to Reconstruction.npz file")

    # FUNCTION TO GET THE TOMOGRAPHY AND RETURN THE DATA
    def __call__(self):
        if self.debug:
            debug_data = {"invert_class": {}}
            start_time = tt.time()
            st = start_time
        # CREATING GEOMETRY MATRIX
        self.geom_matrix()
        # DEBUG TIME
        if self.debug:
            step = "Creating geometry matrix"
            step_time = np.round(tt.time() - st, 2)
            debug_data["invert_class"][step] = step_time
            print(step + ". It took " + str(step_time) + " seconds")
            st = tt.time()
        # CALCULATING TOMOGRAPHY
        self.calc_tomo(
            reg_level=self.reg_level_guess,
            nfisher=self.nfisher,
            eps=self.eps,
            optim_regul=self.optim_regul,
        )

        # DEBUG TIME
        if self.debug:
            step = "Calculating tomography"
            step_time = np.round(tt.time() - st, 2)
            debug_data["invert_class"][step] = step_time
            print(step + ". It took " + str(step_time) + " seconds")
            st = tt.time()

        # FUNCTION TO GET 2D EMISSIVITY DATA
        def get_emissivity_2D(data, eq_data):
            # R AND z VALUES
            R = eq_data["R"]
            z = eq_data["z"]
            rho = eq_data["rho"]
            # EMISSIVITY 1D PROFILE VALUE
            emiss_1D = data["profile"]["sym_emissivity"]
            rho_1D = data["profile"]["rhop"]
            # DATA DECLARATION
            emiss = np.nan * np.ones((len(data["t"]), len(R), len(z)))
            # SWEEP OF TIME
            for it, t in enumerate(data["t"]):
                # SELECTED EQUILIBRIUM INDEX
                sel_ind = np.where(
                    np.abs(eq_data["t"] - t) == np.nanmin(np.abs(eq_data["t"] - t))
                )[0][0]
                # SELECTING RHO
                sel_rho = rho[sel_ind, :, :].T
                # 2D VALUE OF EMISSIVITY
                emiss[it, :, :] = interp1d(
                    rho_1D[it, :], emiss_1D[it, :], bounds_error=False, fill_value=0
                )(sel_rho)
            # RETURN DATA
            return_data = dict(
                R=R,
                z=z,
                data=emiss.T,
            )
            # RETURNING THE DATA
            return return_data

        # GATHERING THE DATA
        return_data = dict(
            t=self.tvec,
            channels_considered=self.valid,  # np.ones(np.size(self.data,1),dtype=bool),
            # BACK INTEGRAL DATA
            back_integral=dict(
                # p_impact=self.impact_parameters,
                data_experiment=self.data,
                data_theory=self.backprojection,
                channel_no=np.arange(1, np.size(self.data, 1) + 1),
            ),
            # PROJECTION DATA
            projection=dict(
                R=self.R,
                z=self.z,
            ),
            # PROFILES
            profile=dict(
                sym_emissivity=self.emiss,
                sym_emissivity_err=self.emiss_err,
                asym_parameter=np.zeros(self.emiss.shape),
                rhop=np.repeat(
                    np.array([self.rho_grid_centers]), len(self.tvec), axis=0
                ),
            ),
        )
        # EMISSIVITY 2D
        return_data["emissivity_2D"] = get_emissivity_2D(return_data, self.eq)
        # ESTIMATING THE CHI2
        data_exp = return_data["back_integral"]["data_experiment"][
            :, return_data["channels_considered"]
        ]
        data_the = return_data["back_integral"]["data_theory"][
            :, return_data["channels_considered"]
        ]
        return_data["back_integral"]["chi2"] = np.sqrt(
            np.nansum(((data_exp - data_the) ** 2) / (data_exp**2), axis=1)
        )
        # APPENDING DEBUG DATA
        if self.debug:
            return_data["debug_data"] = debug_data
        # RETURNING THE DATA
        return return_data



from pathlib import Path
import pickle

from indica import Equilibrium
from indica.configs.readers import ST40Conf
from indica.defaults.load_defaults import get_filename_default_objects
from indica.examples.example_plasma import example_plasma
from indica.operators.atomic_data import default_atomic_data
from indica.readers import ST40Reader

PROJECT_PATH = Path(__file__).parent.parent
DEFAULTS_PATH = f"{PROJECT_PATH}/defaults/"


def save_default_objects(
    machine: str,
    pulse: int,
    tstart: float = 0.02,
    tend: float = 0.1,
    dt: float = 0.01,
    dl: float = 0.005,
    equilibrium_instrument: str = "efit",
):
    """
    Write geometries for specified machine to file for future use as defaults
    """
    if machine == "st40":
        _reader = ST40Reader(pulse, tstart - dt, tend + dt, dt=dt)
        _conf = ST40Conf()
    else:
        raise ValueError(f"Machine {machine} currently not supported")

    plasma_file = get_filename_default_objects(machine)["plasma"]
    equilibrium_file = get_filename_default_objects(machine)["equilibrium"]
    geometry_file = get_filename_default_objects(machine)["geometry"]
    # Diagnostic geometry transform objects
    transforms: dict = {}
    for instr in _conf.INSTRUMENT_METHODS.keys():
        try:
            data = _reader.get("", instr, 0, dl=dl)
            if hasattr(data[list(data)[0]], "transform"):
                _transform = data[list(data)[0]].transform
                transforms[instr] = _transform
        except Exception as e:
            print(f"Error reading {instr}: {e}")
    print(f"\n Writing geometry to: {geometry_file}. \n")
    pickle.dump(transforms, open(geometry_file, "wb"))

    # Equilibrium object
    equilibrium_data = _reader.get("", equilibrium_instrument, 0)
    equilibrium_object = Equilibrium(equilibrium_data)
    print(f"\n Writing equilibrium data to: {equilibrium_file}. \n")
    pickle.dump(equilibrium_object, open(equilibrium_file, "wb"))

    # Plasma object
    plasma = example_plasma(
        machine=machine,
        tstart=tstart,
        tend=tend,
        dt=dt,
        main_ion="h",
        impurities=("c", "ar", "he"),
        impurity_concentration=(0.02, 0.001),  # should be deleted!
        full_run=False,
        n_rad=41,
        n_R=100,
        n_z=100,
    )
    fract_abu, power_loss_tot = default_atomic_data(plasma.elements)
    plasma.fract_abu = fract_abu
    plasma.power_loss_tot = power_loss_tot
    print(f"\n Writing plasma object to: {plasma_file}. \n")
    pickle.dump(plasma, open(plasma_file, "wb"))

    return plasma


if __name__ == "__main__":
    save_default_objects("st40", 11560)



import os

import numpy as np
from scipy.interpolate import CubicSpline
from scipy.interpolate import PchipInterpolator
import yaml

import indica
from indica.profilers.profiler_base import ProfilerBase
from indica.utilities import format_coord
from indica.utilities import format_dataarray


class ProfilerMonoSpline(ProfilerBase):
    def __init__(
        self,
        datatype: str = "electron_temperature",
        coord="rhop",
        xspl: np.ndarray = None,
        parameters: dict = None,
    ):

        super().__init__(parameters)
        self.coord = coord
        self.datatype = datatype

        if xspl is None:
            self.xspl = format_coord(np.linspace(0, 1.0, 30), self.coord)
        else:
            self.xspl = xspl

        if parameters is None:
            self.set_parameters(
                **get_defaults_for_profiler_spline(
                    datatype=datatype, config_name="profiler_monospline"
                )
            )
        else:
            self.set_parameters(**parameters)

        self.x = self.parameters["xknots"]

        if (
            len(self.x) != len(self.parameters) - 1
        ):  # don't count xknot in length of y params
            raise ValueError(
                f"number of y parameters: {len(self.parameters)-1} "
                f"!= number of xknots: {len(self.x)}"
            )

    def __call__(self, *args, **kwargs):

        self.y = [
            _y
            for key, _y in dict(sorted(self.parameters.items())).items()
            if "xknot" not in key
        ]
        self.spline = PchipInterpolator(self.x, self.y)
        _yspl = self.spline(self.xspl)
        coords = {self.coord: self.xspl}
        self.ydata = format_dataarray(_yspl, self.datatype, coords=coords)
        return self.ydata


class ProfilerCubicSpline(ProfilerBase):
    def __init__(
        self,
        datatype: str = "electron_temperature",
        coord="rhop",
        xspl: np.ndarray = None,
        parameters: dict = None,
    ):

        super().__init__(parameters)
        self.coord = coord
        self.datatype = datatype

        if xspl is None:
            self.xspl = format_coord(np.linspace(0, 1.0, 30), self.coord)
        else:
            self.xspl = xspl

        if parameters is None:
            self.set_parameters(
                **get_defaults_for_profiler_spline(
                    datatype=datatype, config_name="profiler_cubicspline"
                )
            )
        else:
            self.set_parameters(**parameters)

        self.x = self.parameters["xknots"]

        if (
            len(self.x) != len(self.parameters) - 1
        ):  # don't count xknot in length of y params
            raise ValueError(
                f"number of y parameters: {len(self.parameters)-1} "
                f"!= number of xknots: {len(self.x)}"
            )

    def __call__(self, *args, **kwargs):

        _yshape = [
            _y
            for key, _y in dict(sorted(self.parameters.items())).items()
            if "shape" in key
        ]

        y = [self.parameters["y0"]]
        for index, shapevalue in enumerate(_yshape):
            y.append(shapevalue * y[index])
        y.append(self.parameters["y1"])

        self.spline = CubicSpline(self.x, y, bc_type="clamped")
        _yspl = self.spline(self.xspl)
        coords = {self.coord: self.xspl}
        self.ydata = format_dataarray(_yspl, self.datatype, coords=coords)
        return self.ydata


def get_defaults_for_profiler_spline(
    datatype="electron_temperature", config_name="profiler_monospline"
):
    """
    Loads config for default parameter values
    """
    path = os.path.join(
        os.path.dirname(indica.__file__), f"configs/profilers/{config_name}.yaml"
    )
    with open(path) as stream:
        cfg = yaml.safe_load(stream)
    return cfg[datatype]



"""Provides implementation of :py:class:`readers.DataReader` for
reading MDS+ data produced by ST40.

"""


from typing import Any
from typing import Dict
from typing import Tuple

import numpy as np

from indica.configs.readers import ST40Conf
from indica.converters import CoordinateTransform
from indica.converters import LineOfSightTransform
from indica.converters import TransectCoordinates
from indica.converters import TrivialTransform
from indica.numpy_typing import RevisionLike
from indica.readers.datareader import DataReader
from indica.readers.mdsutils import MDSUtils


class ST40Reader(DataReader):
    """Class to read ST40 MDS+ data using MDSplus."""

    def __init__(
        self,
        pulse: int,
        tstart: float,
        tend: float,
        machine_conf=ST40Conf,
        reader_utils=MDSUtils,
        server: str = "smaug",
        tree: str = "ST40",
        verbose: bool = False,
        default_error: float = 0.05,
        **kwargs: Any,
    ):

        if tstart < 0:
            tstart = 0

        super().__init__(
            pulse,
            tstart,
            tend,
            machine_conf=machine_conf,
            reader_utils=reader_utils,
            server=server,
            verbose=verbose,
            **kwargs,
        )
        self.default_error = (default_error,)
        self.reader_utils = self.reader_utils(pulse, server, tree)

    def _get_thomson_scattering(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        R = database_results["R"]
        database_results["channel"] = np.arange(len(R))
        database_results["z"] = R * 0.0
        database_results["x"] = R
        database_results["y"] = R * 0.0
        transform = assign_transect_transform(database_results)
        return database_results, transform

    def _get_profile_fits(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        database_results["channel"] = np.arange(len(database_results["R_data"]))
        transform = assign_trivial_transform()
        return database_results, transform

    def _get_charge_exchange(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        database_results["channel"] = np.arange(len(database_results["x"]))
        database_results["element"] = ""
        if "wavelength" in database_results.keys():
            if len(np.shape(database_results["wavelength"])) > 1:
                database_results["wavelength"] = database_results["wavelength"][0, :]
            database_results["pixel"] = np.arange(len(database_results["wavelength"]))

        transform = assign_transect_transform(database_results)
        return database_results, transform

    def _get_spectrometer(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        # Sort channel indexing either hardcore or
        # selecting channels with finite data only
        spectra = database_results["spectra"]
        if database_results["instrument"] == "pi":
            has_data = np.arange(21, 28)
        else:
            has_data = np.where(np.isfinite(spectra[0, :, 0]) * (spectra[0, :, 0] > 0))[
                0
            ]
        database_results["spectra"] = database_results["spectra"][:, has_data, :]
        database_results["spectra_error"] = database_results["spectra_error"][
            :, has_data, :
        ]
        database_results["location"] = database_results["location"][has_data, :]
        database_results["direction"] = database_results["direction"][has_data, :]
        database_results["channel"] = np.arange(database_results["location"][:, 0].size)
        if len(np.shape(database_results["wavelength"])) > 1:
            database_results["wavelength"] = database_results["wavelength"][0, :]

        rearrange_geometry(database_results["location"], database_results["direction"])

        transform = assign_lineofsight_transform(database_results)
        return database_results, transform

    def _get_equilibrium(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        # Add boundary index
        database_results["index"] = np.arange(np.size(database_results["rbnd"][0, :]))
        # Re-shape psi matrix
        database_results["psi"] = database_results["psi"].reshape(
            (
                len(database_results["t"]),
                len(database_results["z"]),
                len(database_results["R"]),
            )
        )
        transform = assign_trivial_transform()
        return database_results, transform

    def _get_radiation(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        database_results["channel"] = np.arange(database_results["location"][:, 0].size)
        transform = assign_lineofsight_transform(database_results)
        return database_results, transform

    def _get_helike_spectroscopy(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        database_results["channel"] = np.arange(database_results["location"][:, 0].size)
        transform = assign_lineofsight_transform(database_results)
        return database_results, transform

    def _get_diode_filters(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        database_results["channel"] = np.arange(database_results["location"][:, 0].size)
        _labels = database_results["label"]
        if type(_labels[0]) == np.bytes_:
            database_results["label"] = np.array(
                [label.decode("UTF-8") for label in _labels]
            )
        else:
            database_results["label"] = _labels
        rearrange_geometry(database_results["location"], database_results["direction"])
        transform = assign_lineofsight_transform(database_results)
        return database_results, transform

    def _get_interferometry(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        # if database_results["instrument"] == "smmh":
        #     location = (location + location_r) / 2.0
        #     direction = (direction + direction_r) / 2.0
        database_results["channel"] = np.arange(database_results["location"][:, 0].size)
        rearrange_geometry(database_results["location"], database_results["direction"])
        transform = assign_lineofsight_transform(database_results)
        return database_results, transform

    def _get_zeff(
        self,
        database_results: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        transform = assign_trivial_transform()
        return database_results, transform

    def __call__(
        self,
        instruments: list = None,
        revisions: Dict[str, RevisionLike] = None,
        debug: bool = False,
    ):

        if instruments is None:
            instruments = self.machine_conf.INSTRUMENT_METHODS.keys()
        if revisions is None:
            revisions = {instrument: 0 for instrument in instruments}
        for instr in instruments:
            if instr not in revisions.keys():
                revisions[instr] = 0

        self.data = {}
        for instrument in instruments:
            print(f"Reading {instrument}")
            try:
                self.data[instrument] = self.get(
                    "",
                    instrument,
                    revisions[instrument],
                )
            except Exception as e:
                print(f"error reading: {instrument} \nException: {e}")
                if debug:
                    raise e
        return self.data


def rearrange_geometry(location, direction):
    if len(np.shape(location)) == 1:
        location = np.array([location])
        direction = np.array([direction])


def assign_lineofsight_transform(database_results: Dict):
    transform = LineOfSightTransform(
        database_results["location"][:, 0],
        database_results["location"][:, 1],
        database_results["location"][:, 2],
        database_results["direction"][:, 0],
        database_results["direction"][:, 1],
        database_results["direction"][:, 2],
        machine_dimensions=database_results["machine_dims"],
        dl=database_results["dl"],
        passes=database_results["passes"],
    )
    return transform


def assign_transect_transform(database_results: Dict):
    transform = TransectCoordinates(
        database_results["x"],
        database_results["y"],
        database_results["z"],
        machine_dimensions=database_results["machine_dims"],
    )

    return transform


def assign_trivial_transform():
    transform = TrivialTransform()
    return transform
